package EV3;

import lejos.hardware.Button; //Zugriff auf die Button Klasse nun moeglich
import lejos.hardware.motor.EV3LargeRegulatedMotor; //Klasse fuer einen grossen EV3-Motor
import lejos.hardware.port.MotorPort; //Zugriff auf die Motorports moeglich (Interface) 
import lejos.hardware.port.SensorPort; //Zugriff auf die Sensorports moeglich (Interface) 
import lejos.hardware.sensor.EV3ColorSensor; //Zugriff auf die Klasse des Farbsensors

/**
 * @author Patric Renner
 * @version 3.0
 * @date 19.06.2018
 * 
 * Klasse fuer einen EV3 Line Follower mit einem Farbsensor und zwei groen Motoren
 * 
 * Idee: Den Roboter auf der rechten Seite der schwarzen Linie aufsetzen und
 * ihn zwischen der schwarzen Linie und dem weissen Bereich entlang fahren lassen.
 * Der Roboter versucht immer auf dieser Linie zu fahren.
 * 
 * Ports: Koennen frei gewaehlt werden, in folgendem Code wurden die Ports
 * A, D fuer die Motoren und Port 1 fuer den Sensor ausgewaehlt. Hier sollten auch
 * die Kabel zu den jeweiligen Ports verbunden sein.
 * 
 * Kalibrierungseinstellungen Zeile 52-94:
 * Hier koennen die Werte veraendert oder angepasst werden.
 * 
 * Empfehlung: Heller Raum und eine Linienbreite von 2 cm.
 * Funktioniert aber auf allen Linienbreiten und auch bei dunklem Raum.
 * 
 * Fuer eine bessere Verstaendlichkeit den Code bitte von oben nach unten lesen.
 * 
 * Weitere Dokus unter http://smart-machines.hs-kl.de/ressourcen-des-ak/ev3/#Dokumentation_und_Tutorials
 * ->Anleitung LeJos Installation "LeJos_Installationsanleitung"
 * ->Anleitung um eine neues Package in Eclipse anzulegen "LeJos_Erste_Schritte"
 * ->Anleitung um ein Programm auf den EV zu Uebertragen "Ein_erstes_Programm_ausfuehren"
 * 
 */
public class LineFollowerEV {

	//Zwei neue Motor Objekte erzeugen:
	//motorA greift auf Port A zu (MotorPort.A)
	//motorD greift auf Port D zu (MotorPort.D)
	private EV3LargeRegulatedMotor motorA = new EV3LargeRegulatedMotor(MotorPort.A);
	private EV3LargeRegulatedMotor motorD = new EV3LargeRegulatedMotor(MotorPort.D);
	
	//Neues ColorSensor Objekt erstellen das auf den Port 1 (S1) zugreift
	private EV3ColorSensor colorSensor = new EV3ColorSensor(SensorPort.S1);
	private float[] farbwerteRGB; 
	
	
	
	//------Kalibrierungseinstellungen: Werte (Konstanten) die veraendert werden koennen------
	//final: Variable kann nur an dieser Stelle verndert werden.
	
	//Der Bereich in dem sich der Roboter bewegen sollte:
	//(Zur Erinnerung der Roboter soll in der Mitte zwischen schwarz und weiss fahren)
	final float WEISS = 0.65f; //Der Roboter befindet sich zu weit rechts vom idealen Wert
	final float MITTE = 0.35f; //Der Roboter befindet sich genau im idealen Wert
	final float SCHWARZ = 0.05f; //Der Roboter befindet sich zu weit links vom idealen Wert
	
	/*Durch den RGB-Farbsensor kann der Roboter auch auf Zwischenwerte reagieren
	 * Es wurde ein Intervall von 0,05 festgelegt. Und bei Tests verfeinert.
	 * 
	 * Je kleiner die Zahl wird, desto groesser ist der Unterschied der Geschwindigkeit
	 * zwischen den zwei Motoren, denn der Roboter befindet sich nicht mehr auf dem idealen
	 * Weg. Je nach Abweichung muss der Roboter staerker oder schwaecher lenken.
	 */
	final float SCAN_A = 0.0451f;
	final float SCAN_B = 0.0951f;
	final float SCAN_C = 0.1451f;
	final float SCAN_D = 0.1951f;
	final float SCAN_E = 0.2451f;
	
	//Hoechstgeschwindigkeit die der Roboter faehrt:
	//(wurde auf 150 festgelegt, da er unseren Parcours so am besten umfahren hat)
	//Tipp: min: -200  (Der EV kann die Raeder auch rueckwaerts drehen)
	//Tipp: max: +200
	final int SPEED = 150;
	
	//Ein Array das die Werte enthaelt die an der Variable Speed abgezogen werden, um die
	//Geschwindigkeit zwischen den Motoren unterschiedlich einzustellen:
	final int[] speedWerte = {
			//In zuWeitRechts erster Wert MotorA, zweiter Wert MotorD
			//In zuWeitLinks erster Wert MotorD, zweiter Wert MotorA
			20, //Speicherort 0 (Hier nur ein Wert da der andere Wert "SPEED" (150) ist)
			20, 50, //Speicherort 1, 2
			40, 90, //Speicherort 3, 4
			70, 145, //Speicherort 5, 6
			90, 170, //Speicherort 7, 8
			90, 200, //Speicherort 9, 10
			80, 220 //Speicherort 11, 12
			}; 
	
	//---------------------Kalibrierungseinstellungen Ende!-------------------

	

	//Der Sensor wird im Konstruktuor in den RGB-Modus (rot, gruen, blau) versetzt
	//Dadurch koennen die Farbwerte genauer ausgelesen werden
	LineFollowerEV() {
		colorSensor.setCurrentMode("RGB");
		
		/* Erzeugen eines Float Arrays zum Speichern der RGB-Werte
		 * Es werden 3 Werte gespeichert, fuer rot, gruen und blau
		 * spaeter werden diese Werte addiert und damit berechnet 
		 * in welchem Bereich sich der Roboter befindet
		 * 
		 * Ueber die Methode sampleSize wird ausgelesen wie viele Werte in colorSensor
		 * enhalten sind und dem Array "farbwerteRGB" eine Groesse von 3 zugewiesen.
		 */
		farbwerteRGB = new float[colorSensor.sampleSize()];
	}
		
	/*
	 * Hauptmethode zum Erzeugen eines Objektes das die Hauptschleife aufruft
	 */		
	public static void main(String[] args) {
		//Erzeugen eines neuen Objektes das die Methode hauptSchleife aufruft:
		LineFollowerEV drive = new LineFollowerEV();
		drive.hauptSchleife();
	}  
	
	
	/*
	 * Zentrale Methode die eine Schleife enthaelt die endlos durchlaufen wird und ermoeglicht
	 * auf die Aenderungen des Farbsensors immer wieder neu zu reagieren
	 */	
	private void hauptSchleife() { 
		//Methode um den EV3 los Fahren zu lassen:
		losFahren();

	     //Solange der Zurueck-Button nicht gedrueckt wird, wird das Programm ausgefuehrt
		 while (Button.ESCAPE.isUp()) {
			 
			//liest den Farbsensor und speichert die RGB-Werte ab Wert 0 im Array ab
			colorSensor.fetchSample(farbwerteRGB, 0); 
	
			//Bilden einer Summe der RGB-Werte um zu ueberpruefen ob sich der Roboter
			//im idealen Bereich bewegt:
			//Auslesen der Werte aus dem Array: farbwerteRGB[...]
		    float summeFarbwerteRGB = farbwerteRGB[0] + farbwerteRGB[1] + farbwerteRGB[2];
	
		    /* Sobald der berechnete Wert aus "summeFarbwerteRGB" hoeher wird befindet man 
		     * sich naeher im weissen Bereich
		     * -> also nach links steuern 
		     * 
		     * Sobald der Wert niedriger wird, befindet man sich naeher im schwarzen Bereich
		     * -> also nach rechts steuern
		     *  
			 *  Weiter koennen die Geschwindigkeiten in den Methoden "zuWeitLinks" "zuWeitRechts" 
			 *  individuell angepasst werden
		     *  
		     */
		    if (summeFarbwerteRGB > WEISS) {  //Zu weit rechts speedWerte 11, 12
		    	zuWeitRechts((SPEED - speedWerte[11]), (SPEED - speedWerte[12]));
		    } else if (summeFarbwerteRGB > (WEISS - SCAN_A)  && summeFarbwerteRGB < WEISS) {  //Zu weit rechts speedWerte 9, 10
		    	zuWeitRechts((SPEED - speedWerte[9]), (SPEED - speedWerte[10]));
		    } else if (summeFarbwerteRGB > (WEISS - SCAN_B) && summeFarbwerteRGB < (WEISS - SCAN_A)) {  //Zu weit rechts speedWerte 7, 8
		    	zuWeitRechts((SPEED - speedWerte[7]), (SPEED - speedWerte[8]));
		    } else if (summeFarbwerteRGB > (WEISS - SCAN_C) && summeFarbwerteRGB < (WEISS - SCAN_B)) {  //Zu weit rechts speedWerte 5, 6
		    	zuWeitRechts((SPEED - speedWerte[5]), (SPEED - speedWerte[6]));
		    } else if (summeFarbwerteRGB > (WEISS - SCAN_D) && summeFarbwerteRGB < (WEISS - SCAN_C)) {  //Zu weit rechts speedWerte 3, 4
		    	zuWeitRechts((SPEED - speedWerte[3]), (SPEED - speedWerte[4]));
		    } else if (summeFarbwerteRGB > (WEISS - SCAN_E) && summeFarbwerteRGB < (WEISS - SCAN_D)) {  //Zu weit rechts speedWerte 1, 2
		    	zuWeitRechts((SPEED - speedWerte[1]), (SPEED - speedWerte[2]));
		    } else if (summeFarbwerteRGB > MITTE && summeFarbwerteRGB < (WEISS - SCAN_E)) { //Mitte SPEED und speedWerte 0
		    	zuWeitRechts(SPEED, (SPEED - speedWerte[0]));
	
			//speedWerte muessen nun umgekehrt eingesetzt werden da der Roboter nun in die andere Richtung nach links fahren soll:
		    } else if (summeFarbwerteRGB > (SCHWARZ + SCAN_E) && summeFarbwerteRGB < MITTE) { //Mitte SPEED und speedWerte 0
		    	zuWeitLinks((SPEED - speedWerte[0]), SPEED);
		    } else if (summeFarbwerteRGB > (SCHWARZ + SCAN_D) && summeFarbwerteRGB < (SCHWARZ + SCAN_E)) { //Zu weit links speedWerte 2, 1
		    	zuWeitLinks((SPEED - speedWerte[2]), (SPEED - speedWerte[1]));
		    } else if (summeFarbwerteRGB > (SCHWARZ + SCAN_C) && summeFarbwerteRGB < (SCHWARZ + SCAN_D)) { //Zu weit links speedWerte 4, 3
		    	zuWeitLinks((SPEED - speedWerte[4]), (SPEED - speedWerte[3]));
		    } else if (summeFarbwerteRGB > (SCHWARZ + SCAN_B) && summeFarbwerteRGB < (SCHWARZ + SCAN_C)) { //Zu weit links speedWerte 6, 5
		    	zuWeitLinks((SPEED - speedWerte[6]), (SPEED - speedWerte[5]));
		    } else if (summeFarbwerteRGB > (SCHWARZ + SCAN_A) && summeFarbwerteRGB < (SCHWARZ + SCAN_B)) { //Zu weit links speedWerte 8, 7
		    	zuWeitLinks((SPEED - speedWerte[8]), (SPEED - speedWerte[7]));
			} else if (summeFarbwerteRGB > SCHWARZ && summeFarbwerteRGB < (SCHWARZ + SCAN_A)) { //Zu weit links speedWerte 10, 9
				zuWeitLinks((SPEED - speedWerte[10]), (SPEED - speedWerte[9]));
		    } else { //Zu weit links speedWerte 12, 11
		    	zuWeitLinks((SPEED - speedWerte[12]), (SPEED - speedWerte[11]));
		    } 
		 }
	}
	
	/*
	 * Methode um den EV3 los fahren zu lassen.
	 */
	public void losFahren () {
		//Die maximal Geschwindigkeit den Motoren zuweisen um loszufahren:
	    motorA.setSpeed(SPEED);
		
	    //Mit der zweiten Methode wird dem Motor 
	    //gesagt, ob er nach vorne oder nach hinten fahren soll:
	    motorA.forward(); //Vor fahren
		
	    //motorA.backward(); //Zurueck fahren

	    //Fr den zweiten Motor das selbe Prinzip:
	    motorD.setSpeed(SPEED);
	    motorD.forward();
	}
	
	
	/*
	 * Methode fuer den weissen Bereich, um zu ueberpruefen ob der Roboter anhand der 
	 * Geschwindigkeit (SPEED) die Motoren nach vorne oder nach hinten drehen lassen muss:
	 */	
	public void zuWeitRechts (int vorA, int zurueckD) {
		motorA.setSpeed(vorA);
		//ist der Wert groeser als 0, Motoren nach vorne drehen
		if (vorA > 0) {
			motorA.forward();
		//ist der Wert kleiner als 0, werden die Motoren zurueck gedreht
		} else {
			motorA.backward();
		}

	    motorD.setSpeed(zurueckD);
		if (zurueckD > 0) {
			motorD.forward();
		} else {
			motorD.backward();
		}
	}
	
	/*
	 * Methode fuer den schwarzen Bereich, um zu ueberpruefen ob der Roboter anhand der 
	 * Geschwindigkeit (SPEED) die Motoren nach vorne oder nach hinten drehen lassen muss:
	 */
	public void zuWeitLinks (int zurueckA, int vorD) {
		motorA.setSpeed(zurueckA);
		if (zurueckA > 0) {
			motorA.forward();
		} else {
			motorA.backward();
		}
		
	    motorD.setSpeed(vorD);
		if (vorD > 0) {
			motorD.forward();
		} else {
			motorD.backward();
		}
	}	
}